@@ -1,19 +1,14 @@
-# This represents one _session_ in a Chromote object. Note that in the Chrome
-# DevTools Protocol a session is a debugging interface connected to a
-# _target_; a target is a browser window/tab, or an iframe. A single target
-# can have more than one session connected to it.
-
 #' ChromoteSession class
+#'
+#' @description
+#' This represents one _session_ in a Chromote object. Note that in the Chrome
+#' DevTools Protocol a session is a debugging session connected to a _target_,
+#' which is a browser window/tab or an iframe.
+#'
+#' A single target can potentially have more than one session connected to it,
+#' but this is not currently supported by chromote.
+#'
 #' @export
-#' @param timeout_ Number of seconds for \pkg{chromote} to wait for a Chrome
-#' DevTools Protocol response. If `timeout_` is [`rlang::missing_arg()`] and
-#' `timeout` is provided, `timeout_` will be set to `2 * timeout / 1000`.
-#' @param timeout Number of milliseconds for Chrome DevTools Protocol execute a
-#' method.
-#' @param width Width, in pixels, of the `Target` to create if `targetId` is
-#'   `NULL`
-#' @param height Height, in pixels, of the `Target` to create if `targetId` is
-#'   `NULL`
 #' @param targetId
 #'   [Target](https://chromedevtools.github.io/devtools-protocol/tot/Target/)
 #'   ID of an existing target to attach to. When a `targetId` is provided, the
@@ -49,6 +44,7 @@ ChromoteSession <- R6Class(
     #'   from the parent `Chromote` object. If `TRUE`, enable automatic
     #'   event enabling/disabling; if `FALSE`, disable automatic event
     #'   enabling/disabling.
+    #' @param width,height Width and height of the new window.
     #' @param wait_ If `FALSE`, return a [promises::promise()] of a new
     #'   `ChromoteSession` object. Otherwise, block during initialization, and
     #'   return a `ChromoteSession` object directly.
@@ -76,9 +72,11 @@ ChromoteSession <- R6Class(
           wait_ = FALSE
          )$
           then(function(value) {
+            private$target_id <- value$targetId
             parent$Target$attachToTarget(value$targetId, flatten = TRUE, wait_ = FALSE)
           })
       } else {
+        private$target_id <- targetId
         p <- parent$Target$attachToTarget(targetId, flatten = TRUE, wait_ = FALSE)
       }
 
@@ -100,7 +98,8 @@ ChromoteSession <- R6Class(
 
       private$auto_events <- auto_events
       private$event_manager <- EventManager$new(self)
-      private$is_active_ <- TRUE
+      private$session_is_active <- TRUE
+      private$target_is_active <- TRUE
 
       # Find pixelRatio for screenshots
       p <- p$
@@ -119,7 +118,7 @@ ChromoteSession <- R6Class(
               warning("Chromote has received a Inspector.targetCrashed event. This means that the ChromoteSession has probably crashed.")
               # Even if no targetId nor sessionId is returned by Inspector.targetCashed
               # mark the session as closed. This will close all sessions..
-              self$mark_closed()
+              self$mark_closed(TRUE)
             })
           })
       }
@@ -131,7 +130,7 @@ ChromoteSession <- R6Class(
         # returning p, because the call to ChromoteSession$new() always
         # returns the new object. Instead, we'll store it as
         # private$init_promise_, and the user can retrieve it with
-        # b$init_promise().
+        # b$get_init_promise().
         private$init_promise_ <- p$then(function(value) self)
       }
 
@@ -157,11 +156,9 @@ ChromoteSession <- R6Class(
     #' if (interactive()) b$view()
     #' ```
     view = function() {
-      tid <- self$Target$getTargetInfo()$targetInfo$targetId
-
       # A data frame of targets, one row per target.
       info <- fromJSON(self$parent$url("/json"))
-      path <- info$devtoolsFrontendUrl[info$id == tid]
+      path <- info$devtoolsFrontendUrl[info$id == private$target_id]
       if (length(path) == 0) {
         stop("Target info not found.")
       }
@@ -188,18 +185,19 @@ ChromoteSession <- R6Class(
     #' when the `ChromoteSession` is closed. Otherwise, block until the
     #' `ChromoteSession` has closed.
     close = function(wait_ = TRUE) {
-      p <- self$Target$getTargetInfo(wait_ = FALSE)
-      p <- p$then(function(target) {
-        tid <- target$targetInfo$targetId
-        # Even if this session calls Target.closeTarget, the response from
-        # the browser is sent without a sessionId. In order to wait for the
-        # correct browser response, we need to invoke this from the parent's
-        # browser-level methods.
-        self$parent$protocol$Target$closeTarget(tid, wait_ = FALSE)
-      })
+      if (!private$target_is_active) {
+        return(invisible())
+      }
+
+      # Even if this session calls Target.closeTarget, the response from
+      # the browser is sent without a sessionId. In order to wait for the
+      # correct browser response, we need to invoke this from the parent's
+      # browser-level methods.
+      p <- self$parent$protocol$Target$closeTarget(private$target_id, wait_ = FALSE)
+
       p <- p$then(function(value) {
         if (isTRUE(value$success)) {
-          self$mark_closed()
+          self$mark_closed(TRUE)
         }
         invisible(value$success)
       })
@@ -286,7 +284,12 @@ ChromoteSession <- R6Class(
     #' #> Done!
     #' ```
     #'
-    #' @param filename File path of where to save the screenshot.
+    #' @param filename File path of where to save the screenshot. The format of
+    #'   the screenshot is inferred from the file extension; use
+    #'   `options = list(format = "jpeg")` to manually choose the format. See
+    #'   [`Page.captureScreenshot`](https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-captureScreenshot)
+    #'   for supported formats; at the time of this release the format options
+    #'   were `"png"` (default), `"jpeg"`, or `"webp"`.
     #' @param selector CSS selector to use for the screenshot.
     #' @param cliprect A list containing `x`, `y`, `width`, and `height`. See
     #' [`Page.Viewport`](https://chromedevtools.github.io/devtools-protocol/tot/Page/#type-Viewport)
@@ -300,6 +303,8 @@ ChromoteSession <- R6Class(
     #' @param delay The number of seconds to wait before taking the screenshot
     #' after resizing the page. For complicated pages, this may need to be
     #' increased.
+    #' @param options Additional options passed to
+    #'   [`Page.captureScreenshot`](https://chromedevtools.github.io/devtools-protocol/tot/Page/#method-captureScreenshot).
     #' @param wait_ If `FALSE`, return a [promises::promise()] that will resolve
     #' when the `ChromoteSession` has saved the screenshot. Otherwise, block
     #' until the `ChromoteSession` has saved the screenshot.
@@ -312,6 +317,7 @@ ChromoteSession <- R6Class(
       scale = 1,
       show = FALSE,
       delay = 0.5,
+      options = list(),
       wait_ = TRUE
     ) {
       chromote_session_screenshot(
@@ -324,6 +330,7 @@ ChromoteSession <- R6Class(
         scale = scale,
         show = show,
         delay = delay,
+        options = options,
         wait_ = wait_
       )
     },
@@ -407,23 +414,43 @@ ChromoteSession <- R6Class(
     #' when the `ChromoteSession` has created a new session. Otherwise, block
     #' until the `ChromoteSession` has created a new session.
     new_session = function(width = 992, height = 1323, targetId = NULL, wait_ = TRUE) {
-      self$parent$new_session(width = width, height = height, targetId = targetId, wait_ = wait_)
+      create_session(
+        chromote = self$parent,
+        width = width,
+        height = height,
+        targetId = targetId,
+        wait_ = wait_
+      )
     },
 
     #' @description
     #' Retrieve the session id
-    #'
-    #' ## Examples
-    #'
-    #' ```r
-    #' b <- ChromoteSession$new()
-    #' b$get_session_id()
-    #' #> [1] "05764F1D439F4292497A21C6526575DA"
-    #' ```
     get_session_id = function() {
       private$session_id
     },
 
+    #' @description
+    #' Create a new session that connects to the same target (i.e. page)
+    #' as this session. This is useful if the session has been closed but the target still
+    #' exists.
+    respawn = function() {
+      if (!private$target_is_active) {
+        stop("Can't respawn session; target has been closed.")
+      }
+
+      create_session(
+        chromote = self$parent,
+        targetId = private$target_id,
+        auto_events = private$auto_events
+      )
+    },
+
+    #' @description
+    #' Retrieve the target id
+    get_target_id = function() {
+      private$target_id
+    },
+
     #' @description
     #' Wait for a Chromote Session to finish. This method will block the R
     #' session until the provided promise resolves. The loop from
@@ -485,9 +512,7 @@ ChromoteSession <- R6Class(
     #' @param timeout Number of milliseconds for Chrome DevTools Protocol
     #' execute a method.
     send_command = function(msg, callback = NULL, error = NULL, timeout = NULL) {
-      if (!private$is_active_) {
-        stop("Session ", private$session_id, " is closed.")
-      }
+      self$check_active()
       self$parent$send_command(msg, callback, error, timeout, sessionId = private$session_id)
     },
 
@@ -513,28 +538,72 @@ ChromoteSession <- R6Class(
       private$event_manager$invoke_event_callbacks(event, params)
     },
 
-    #' @description
-    #' Disable callbacks for a given session.
-    #'
-    #' For internal use only.
-    mark_closed = function() {
-      private$is_active_ <- FALSE
+    #' @description Mark a session, and optionally, the underlying target,
+    #'   as closed. For internal use only.
+    #' @param target_closed Has the underlying target been closed as well as the
+    #'   active debugging session?
+    mark_closed = function(target_closed) {
+      private$session_is_active <- FALSE
+      private$target_is_active <- !target_closed
     },
 
     #' @description Retrieve active status
     #' Once initialized, the value returned is `TRUE`. If `$close()` has been
     #' called, this value will be `FALSE`.
     is_active = function() {
-      private$is_active_
+      private$session_is_active && private$target_is_active && self$parent$is_active()
+    },
+
+    #' @description Check that a session is active, erroring if not.
+    check_active = function() {
+      if (self$is_active()) {
+        return()
+      }
+
+      if (private$target_is_active) {
+        abort(c(
+          "Session has been closed.",
+          i = "Call session$respawn() to create a new session that connects to the same target."
+        ))
+      } else {
+        abort("Session and underlying target have been closed.")
+      }
     },
 
     #' @description Initial promise
     #'
     #' For internal use only.
-    init_promise = function() {
+    get_init_promise = function() {
       private$init_promise_
     },
 
+    #' @description Summarise the current state of the object.
+    #' @param verbose The print method defaults to a brief summary
+    #'   of the most important debugging info; use `verbose = TRUE` tp
+    #'   see the complex R6 object.
+    #' @param ... Passed on to `format()` when `verbose` = TRUE
+    print = function(..., verbose = FALSE) {
+      if (verbose) {
+        cat(format(self, ...), sep = "\n")
+      } else {
+        if (self$is_active()) {
+          state <- "session + target active"
+        } else if (private$target_is_active) {
+          state <- "target active"
+        } else {
+          state <- "closed"
+        }
+
+        cat_line("<ChromoteSession> (", state, ")")
+        if (self$is_active())
+          cat_line("  Session ID: ", self$get_session_id())
+        if (private$target_is_active)
+          cat_line("  Target ID:  ", self$get_target_id())
+        cat_line("  Parent PID: ", self$parent$get_browser()$get_process()$get_pid())
+      }
+      invisible(self)
+    },
+
     #' @field parent [`Chromote`] object
     parent = NULL,
     #' @field default_timeout Default timeout in seconds for \pkg{chromote} to
@@ -546,17 +615,44 @@ ChromoteSession <- R6Class(
 
   private = list(
     session_id = NULL,
-    is_active_ = NULL,
+    target_id = NULL,
+    session_is_active = NULL,
+    target_is_active = NULL,
     event_manager = NULL,
     pixel_ratio = NULL,
     auto_events = NULL,
     init_promise_ = NULL,
 
     register_event_listener = function(event, callback = NULL, timeout = NULL) {
-      if (!private$is_active_) {
-        stop("Session ", private$session_id, " is closed.")
-      }
+      self$check_active()
       private$event_manager$register_event_listener(event, callback, timeout)
     }
   )
 )
+
+
+# Wrapper around ChromoteSession$new() that can return a promise
+create_session <- function(chromote = default_chromote_object(),
+                           width = 992,
+                           height = 1323,
+                           targetId = NULL,
+                           wait_ = TRUE,
+                           auto_events = NULL) {
+
+  session <- ChromoteSession$new(
+    parent = chromote,
+    width = width,
+    height = height,
+    targetId,
+    auto_events = auto_events,
+    wait_ = wait_
+  )
+
+  if (wait_) {
+    session
+  } else {
+    # ChromoteSession$new() must return a ChromoteSession object so we need a
+    # side-channel to return a promise
+    session$get_init_promise()
+  }
+}
