@@ -1,3 +1,7 @@
+cat_line <- function(...) {
+  cat(paste0(..., "\n", collapse = ""))
+}
+
 # =============================================================================
 # System
 # =============================================================================
